package problem;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Stream;

import problem.fdr.FDROperator;
import problem.fdr.FDRPartialState;
import problem.fdr.FDRProblem;
import problem.fdr.FDRVar;
import problem.strips.StripsOperator;
import problem.strips.StripsProblem;

public class Parser {
	
	private enum FDRParserState{
		NUM_OF_VARIABLES,NUM_OF_VALUES,VALUES,INIT,GOAL,NUM_OF_OPERATORS,OP_NAME,OP_PRE,OP_EFF,OP_COST
	}
	
	private enum StripsParserState{
		NUM_OF_FACTS,FACTS,INIT,GOAL,NUM_OF_OPERATORS,OP_NAME,OP_PRE,OP_ADD,OP_DEL,OP_COST
	}
	
	
	private FDRPartialState parsePartialState(int[] tokens) {
		
		Map<Integer,Integer> values = new HashMap<>();
		
		for(int i = 0; i < tokens[0]; ++i) {
			values.put(tokens[2*i+1], tokens[2*i+2]);
		}
		
		
		return new FDRPartialState(values);
	}

	public FDRProblem parseFDRProblem(String filename) {
		FDRParserState parseState = FDRParserState.NUM_OF_VARIABLES;
		
		try {
			File file = new File(filename);
			FileReader fileReader = new FileReader(file);
			BufferedReader bufferedReader = new BufferedReader(fileReader);
			
			String line;
			int nof_variables=0;
			int nof_values = 0;
//			int nof_operators = 0;
			List<String> currentValNames = new LinkedList<>();
			List<FDRVar> variables = new LinkedList<>();
			FDRPartialState init = null;
			FDRPartialState goal = null;
			String op_name = null;
			FDRPartialState pre = null;
			FDRPartialState eff = null;
			List<FDROperator> operators = new LinkedList<>();
			
			while ((line = bufferedReader.readLine()) != null) {
				
				switch(parseState) {
					case NUM_OF_VARIABLES:
						nof_variables = Integer.parseInt(line.trim());
						parseState = FDRParserState.NUM_OF_VALUES;
						break;
					case NUM_OF_VALUES:
						nof_values = Integer.parseInt(line.trim());
						parseState = FDRParserState.VALUES;
						break;
					case VALUES:
						currentValNames.add(line.trim());
						nof_values--;
						if(nof_values<=0) {
							variables.add(new FDRVar(variables.size(),currentValNames));
							currentValNames.clear();
							nof_variables--;
							if(nof_variables <= 0) {
								parseState = FDRParserState.INIT;
							}else {
								parseState = FDRParserState.NUM_OF_VALUES;
							}
						}
						break;
					case INIT:
						String[] tokens = line.trim().split(" ");
						int[] values = Stream.of(tokens).mapToInt(Integer::parseInt).toArray();
						init = new FDRPartialState(values);
						parseState = FDRParserState.GOAL;
						break;
					case GOAL:
						goal = parsePartialState(Stream.of(line.trim().split(" ")).mapToInt(Integer::parseInt).toArray());
						parseState = FDRParserState.NUM_OF_OPERATORS;
						break;
					case NUM_OF_OPERATORS:
//						nof_operators = Integer.parseInt(line.trim());
						parseState = FDRParserState.OP_NAME;
						break;
					case OP_NAME:
						op_name = line.trim();
						parseState = FDRParserState.OP_PRE;
						break;
					case OP_PRE:
						pre = parsePartialState(Stream.of(line.trim().split(" ")).mapToInt(Integer::parseInt).toArray());
						parseState = FDRParserState.OP_EFF;
						break;
					case OP_EFF:
						eff = parsePartialState(Stream.of(line.trim().split(" ")).mapToInt(Integer::parseInt).toArray());
						parseState = FDRParserState.OP_COST;
						break;
					case OP_COST:
						operators.add(new FDROperator("("+op_name+")", pre, eff, Integer.parseInt(line.trim())));
						parseState = FDRParserState.OP_NAME;
						break;
					default:
						throw new RuntimeException("Bad parse state");
				}
			}
			fileReader.close();
			
			return new FDRProblem(variables.toArray(new FDRVar[0]), init, goal, operators);
			
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return null;
	}
	
	private Set<Integer> parseFacts(String[] tokens){
		Set<Integer> set = new HashSet<>();
		
		for(int i = 1; i < tokens.length ; ++i) {
			set.add(Integer.parseInt(tokens[i]));
		}
		
		return set;
	}
	
	
	public StripsProblem parseStripsProblem(String filename) {
		StripsParserState parseState = StripsParserState.NUM_OF_FACTS;
		
		try {
			File file = new File(filename);
			FileReader fileReader = new FileReader(file);
			BufferedReader bufferedReader = new BufferedReader(fileReader);
			
			String line;
			int nof_facts=0;
//			int nof_operators = 0;
			ArrayList<String> facts = new ArrayList<>();
			Set<Integer> init = new HashSet<>();
			Set<Integer> goal = new HashSet<>();
			String op_name = null;
			Set<Integer> pre = new HashSet<>();
			Set<Integer> add = new HashSet<>();
			Set<Integer> del = new HashSet<>();
			List<StripsOperator> operators = new LinkedList<>();
			
			while ((line = bufferedReader.readLine()) != null) {
				
				switch(parseState) {
					case NUM_OF_FACTS:
						nof_facts = Integer.parseInt(line.trim());
						parseState = StripsParserState.FACTS;
						break;
					case FACTS:
						facts.add(line.trim());
						nof_facts--;
						if(nof_facts<=0) {
							parseState = StripsParserState.INIT;
						}
						break;
					case INIT:
						String[] tokens = line.trim().split(" ");
						init = parseFacts(tokens);
						parseState = StripsParserState.GOAL;
						break;
					case GOAL:
						tokens = line.trim().split(" ");
						goal = parseFacts(tokens);
						parseState = StripsParserState.NUM_OF_OPERATORS;
						break;
					case NUM_OF_OPERATORS:
//						nof_operators = Integer.parseInt(line.trim());
						parseState = StripsParserState.OP_NAME;
						break;
					case OP_NAME:
						op_name = line.trim();
						parseState = StripsParserState.OP_PRE;
						break;
					case OP_PRE:
						tokens = line.trim().split(" ");
						pre = parseFacts(tokens);
						parseState = StripsParserState.OP_ADD;
						break;
					case OP_ADD:
						tokens = line.trim().split(" ");
						add = parseFacts(tokens);
						parseState = StripsParserState.OP_DEL;
						break;
					case OP_DEL:
						tokens = line.trim().split(" ");
						del = parseFacts(tokens);
						parseState = StripsParserState.OP_COST;
						break;
					case OP_COST:
						operators.add(new StripsOperator("("+op_name+")", pre, add,del, Integer.parseInt(line.trim())));
						parseState = StripsParserState.OP_NAME;
						break;
					default:
						throw new RuntimeException("Bad parse state");
				}
			}
			fileReader.close();
			
			return new StripsProblem(facts, init, goal,operators);
			
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return null;
	}

	
	
	

}
